/*++
Copyright (c) 2011 Microsoft Corporation

Module Name:

    z3_log_frontend.h

Abstract:

    Z3 log frontend.
    Replay a log generated by Z3

Author:

    Leonardo de Moura (leonardo) 2011-09-26.

Revision History:

--*/
#pragma once

void replay_z3_log(char const * benchmark_file);



