/*++
Copyright (c) 2017 Microsoft Corporation

Module Name:

    solver2lookahead.cpp

Abstract:

    Lookahead wrapper for arbitrary solver.

Author:

    Nikolaj Bjorner (nbjorner) 2017-10-9

Notes:
   
--*/
#include "sat/sat_solver/inc_sat_solver.h"
#include "solver/solver.h"

solver * mk_solver2lookahead(solver* s) {
    return nullptr;
}
