/*++
Copyright (c) 2012 Microsoft Corporation

Module Name:

    Deprecated.cs

Abstract:

    Expose deprecated features for use from the managed API 
    those who use them for experiments.

Author:

    Christoph Wintersteiger (cwinter) 2012-03-15

Notes:
    
--*/
using System.Diagnostics;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Microsoft.Z3
{
    /// <summary>
    /// The main interaction with Z3 happens via the Context.
    /// </summary>
    public class Deprecated 
    {


    }
}