/*++
Copyright (c) 2011 Microsoft Corporation

Module Name:

    dimacs_frontend.h

Abstract:

    <abstract>

Author:

    Leonardo de Moura (leonardo) 2011-07-26.

Revision History:

--*/
#pragma once

unsigned read_dimacs(char const * benchmark_file);


