/*++
Copyright (c) 2011 Microsoft Corporation

Module Name:

    polynomial_cmds.h

Abstract:
    Commands for debugging polynomial module.

Author:

    Leonardo (leonardo) 2011-12-23

Notes:

--*/
#pragma once

class cmd_context;
void install_polynomial_cmds(cmd_context & ctx);

