/*++
Copyright (c) 2015 Microsoft Corporation

Module Name:

    automaton.cpp

Abstract:

    Symbolic Automaton, a la Margus Veanes Automata library.

Author:

    Nikolaj Bjorner (nbjorner) 2015-12-23.

Revision History:


--*/

#include "math/automata/automaton.h"

template class automaton<unsigned>;
